@@ -3,8 +3,8 @@ use warnings;
 
 package SVN::Core;
 use SVN::Base qw(Core svn_ VERSION);
-$SVN::Core::VERSION = "$SVN::Core::VER_MAJOR.$SVN::Core::VER_MINOR." .
-                      "$SVN::Core::VER_MICRO";
+# Some build tool hates VERSION assign across two lines.
+$SVN::Core::VERSION = "$SVN::Core::VER_MAJOR.$SVN::Core::VER_MINOR.$SVN::Core::VER_MICRO";
 
 =head1 NAME
 
@@ -48,9 +48,11 @@ functions.
 
 BEGIN {
     SVN::_Core::apr_initialize();
-
 }
 
+our $gpool = SVN::Pool->new_default;
+SVN::Core::utf_initialize($gpool);
+
 END {
     SVN::_Core::apr_terminate();
 }
@@ -199,7 +201,7 @@ sub getline
     my $self = shift;
     *$self->{pool} ||= SVN::Core::pool_create (undef);
     my ($buf, $eof) = *$self->{svn_stream}->readline ($/, *$self->{pool});
-    return undef if $eof && !$buf;
+    return undef if $eof && !length($buf);
     return $eof ? $buf : $buf.$/;
 }
 
@@ -217,14 +219,15 @@ sub READLINE
     my $self = shift;
     unless (defined $/) {
 	my $buf = '';
-	while (my $chunk = *$self->{svn_stream}->read
-	       ($SVN::Core::STREAM_CHUNK_SIZE)) {
+	while (length( my $chunk = *$self->{svn_stream}->read
+	       ($SVN::Core::STREAM_CHUNK_SIZE)) ) {
 	    $buf .= $chunk;
 	}
 	return $buf;
     }
     elsif (ref $/) {
-        return *$self->{svn_stream}->read (${$/}) || undef;
+        my $buf = *$self->{svn_stream}->read (${$/});
+	return length($buf) ? $buf : undef;
     }
     return wantarray ? $self->getlines : $self->getline;
 }
@@ -923,6 +926,18 @@ do not handle yet.
 package SVN::Auth::SSL;
 use SVN::Base qw(Core SVN_AUTH_SSL_);
 
+package _p_svn_lock_t;
+use SVN::Base qw(Core svn_lock_t_);
+
+package SVN::MD5;
+use overload
+    '""' => sub { SVN::Core::md5_digest_to_cstring(${$_[0]})};
+
+sub new {
+    my ($class, $digest) = @_;
+    bless \$digest, $class;
+}
+
 =head1 AUTHORS
 
 Chia-liang Kao E<lt>clkao@clkao.orgE<gt>
